package net.lemnik.eodsql;

import java.lang.annotation.Target;
import java.lang.annotation.Retention;
import java.lang.annotation.ElementType;
import java.lang.annotation.RetentionPolicy;

import java.sql.ResultSet;
import java.sql.Statement;

import net.lemnik.eodsql.spi.MethodImplementation;
import net.lemnik.eodsql.spi.util.NoDataObjectBinding;
import net.lemnik.eodsql.spi.util.DataObjectBinding;

/**
 * <p>
 *    The <code>@Select</code> annotation is used to mark methods within a
 *    {@link BaseQuery Query} class that will return a <code>DataSet</code>.
 *    The <code>Select</code> annotation contains the SQL query that will be
 *    executed on the database, and whether the resulting {@link DataSet} will
 *    be a "connected" or {@link #disconnected "disconnected"}
 *    <code>DataSet</code> to start with. Remember that a <code>DataSet</code>
 *    may be cleanly {@link DataSet#disconnect disconnected} from the database.
 * </p><p>
 *    Here is an example of how a <code>@Select</code> is used:
 *    
 *    <pre>
 *        public class User {
 *            public Integer id;
 *            public String username;
 *        }
 *        
 *        public interface UserQuery extends BaseQuery {
 *            <span style="color: #00f;">@Select("SELECT * FROM user WHERE id = ?1")</span>
 *            public DataSet&lt;User&gt; getUserById(int id);
 *        }
 *    </pre>
 *    
 *    Any method annotated with <code>@Select</code> may return any one of the
 *    following datatypes:
 *    
 *    <ul>
 *        <li>
 *            the exact class to map - this will either return the first result
 *            found, or <code>null</code> if no results where returned.
 *        </li><li>
 *            the class wrapped in a {@link DataSet} - this is the preferred
 *            way of returning multiple results, as it can remain connected
 *            to the database, or be {@link DataSet#disconnect disconnected}
 *            at any time.
 *        </li><li>
 *            an array of the class to bind to
 *        </li><li>
 *            the class wrapped in a {@link java.util.Set},
 *            {@link java.util.List} or {@link java.util.Collection} -
 *            these types have no connection to the database, the data is
 *            loaded into them and the result is returned.
 *        </li>
 *        
 *        Don't forget that {@link DataSet} inherits from {@link java.util.List},
 *        and has all of it's read functionality.
 *    </ul>
 * </p><p>
 *    Columns that cannot be mapped to fields are
 *    ignored, as are fields that cannot be mapped to columns. <code>@Select</code>
 *    SQL uses the exact same query parser as the {@link Update} annotation, and
 *    so can also accept complex types as it's parameters, and insert their
 *    fields into the SQL query.
 * </p>
 * @author jason
 * @see Update
 */
@Target(ElementType.METHOD)
@Retention(RetentionPolicy.RUNTIME)
public @interface Select {
    /**
     * The SQL query to execute on the database. The result set of the query will
     * be mapped to the datatype specified in the method declaration.
     * @see #sql
     */
    String value() default "";

    /**
     * The SQL query to execute on the database. The result set of the query will
     * be mapped to the datatype specified in the method declaration. This value
     * works in the exact same way as {@link #value}, but is used when you wish
     * to specify values other than the SQL query.
     */
    String sql() default "";

    /**
     * Specify whether or not the {@link DataSet} resulting from a call to the
     * annotated method will be connected to the database or not.
     * @see DataSet
     * @see DataSet#disconnect
     */
    boolean disconnected() default false;

    /**
     * If <code>rubberstamp</code> is flagged as <code>true</code>, the annotated
     * method must return a {@link DataIterator}, otherwise it will be rejected by
     * the {@link QueryTool}. Rubber-stamping will cause the returned
     * <code>DataIterator</code> to use a single instance of the data-object for all
     * of the rows returned, instead of initializing a new one for each row. This can
     * be used to save CPU time and memory, and is especially useful in display code.
     */
    boolean rubberstamp() default false;

    /**
     * <p>
     * By deafult a {@link DataSet} returned by an <code>@Select</code> method is
     * read only. If you set this to <code>false</code> the returned <code>DataSet</code>
     * will implement the following additional methods:
     * </p>
     * <ul>
     *  <li>{@link java.util.List#add(Object)}</li>
     *  <li>{@link java.util.List#addAll(java.util.Collection)}</li>
     *  <li>{@link java.util.List#remove(int)}</li>
     *  <li>{@link java.util.List#set(int, Object)}</li>
     * </ul>
     * <p>
     * It is considered a validation error for a method to have readOnly set <code>false</code>
     * if it doesn't return a <code>DataSet</code>.
     * </p>
     *
     * @since 1.1
     */
    boolean readOnly() default true;

    /**
     *<p>
     * The cache class to be used by <code>DataSet</code>'s returned by the
     * annotated method. This only has an effect of methods that return a
     * {@link DataSet} type.
     *</p><p>
     * The default <code>DataSetCache</code> implementation is {@link ArrayDataSetCache},
     * which uses an array to cache the row values. For larger <code>ResultSet</code>'s, you
     * may want to look into a {@link NullDataSetCache} or using a {@link DataIterator} instead of
     * the <code>DataSet</code> class.
     *</p>
     */
    Class<? extends DataSetCache> cache() default ArrayDataSetCache.class;

    /**
     * <p>
     * This is a hint to the database driver to suggest the number of rows
     * to fetch in a single page of the {@link ResultSet}. This value
     * directly corresponds to the {@link Statement#setFetchSize(int)}
     * method.
     * </p><p>
     * If this value is {@literal 0} (which it is by default) then the
     * fetch-size of the underlying {@code Statement} will be left alone.
     * This value (like the {@code fetch-size} of a {@code Statement}) is
     * to be considered a hint to EoD SQL. The underlying
     * {@link MethodImplementation} may choose to ignore this value.
     * </p>
     *
     * @since 2.1
     */
    int fetchSize() default 0;

    /**
     * <p>
     * The custom type bindings to be used on the parameters of this query. The binding is matched 
     * by position. The default type binding will be used for each parameter that has no  
     * <code>parameterBindings</code> entry or an entry that is of type {@link TypeMapper} itself. 
     * </p>
     * 
     * @since 2.1
     */
    Class<? extends TypeMapper>[] parameterBindings() default {};
    
    /**
     * <p>
     * A custom data object binding to be used on the result set of this query. The default value
     * {@link NoDataObjectBinding} indicates that no custom binding is to be used.
     * </p>
     * 
     * @since 2.1
     */
    Class<? extends DataObjectBinding> resultSetBinding() default NoDataObjectBinding.class;
}
